@@ -1,9 +1,11 @@
 from enum import Enum
-import datetime
+import datetime, math
+
 import json
 import re
 import sys
 import requests
+import pprint
 from typing import Optional, Sequence, Union, List
 
 from rspace_client.client_base import ClientBase, Pagination
@@ -33,7 +35,7 @@ class FillingStrategy(Enum):
 
 class Sample:
     """
-     Wraps a dict of Sample data returned from samples/{id} GET API call
+    Wraps a dict of Sample data returned from samples/{id} GET API call
     """
 
     def __init__(self, data: dict):
@@ -41,7 +43,7 @@ class Sample:
 
     def wherep(self) -> List[str]:
         """
-        Returns a list of breadcrumb names of all containers that subsamples 
+        Returns a list of breadcrumb names of all containers that subsamples
         of this sample are located in.
 
         Returns
@@ -55,13 +57,13 @@ class Sample:
             bcumbs.add(b_crumb)
         return bcumbs
 
-    def __str__(self):
+    def __repr__(self):
         return f"Sample: id = {self.data['id']}, name = {self.data['name']}, creationDate = {self.data['created']}"
 
 
 class GridPlacement:
     """
-     Superclass of all grid placement strategies
+    Superclass of all grid placement strategies
     """
 
     def __init__(self, items_to_move: str, filling_strategy: FillingStrategy):
@@ -77,7 +79,7 @@ class GridPlacement:
 
 class AutoFit(GridPlacement):
     """
-     Base class of ByRow and ByColumn filling strategies.
+    Base class of ByRow and ByColumn filling strategies.
     """
 
     def __init__(
@@ -105,6 +107,12 @@ class AutoFit(GridPlacement):
         self.total_columns = total_columns
         self.total_rows = total_rows
 
+    def __repr__(self):
+        return f"""<{self.__class__.__name__}: Items {len(self.items_to_move)}, column_index={self.column_index}
+                row_index = {self.row_index}, total_columns={self.total_columns}, total_rows={self.total_rows},
+                filling_strategy = {self.filling_strategy!r}
+                """
+
 
 class GridLocation:
     """
@@ -117,11 +125,19 @@ class GridLocation:
         self.x = x
         self.y = y
 
+    def __repr__(self):
+        return f"{self.__class__.__name__}({self.x}, {self.y})"
+
+    def __eq__(self, other):
+        if not isinstance(other, self.__class__):
+            return False
+        return self.x == other.x and self.y == other.y
+
 
 class ByRow(AutoFit):
     """
-      Defines a strategy for filling a grid container with a list of items, filling rows
-      in turn, from a starting location. 
+    Defines a strategy for filling a grid container with a list of items, filling rows
+    in turn, from a starting location.
     """
 
     def __init__(
@@ -144,7 +160,7 @@ class ByRow(AutoFit):
             The total number of columns in the grid
         total_rows : int
             The total number of rows in the grid
-        *items_to_move : 
+        *items_to_move :
             One or more global Ids.
          : TYPE
             DESCRIPTION.
@@ -166,8 +182,8 @@ class ByRow(AutoFit):
 
 class ByColumn(AutoFit):
     """
-      Defines a strategy for filling a grid container with a list of items, filling columns
-      in turn, from a starting location. 
+    Defines a strategy for filling a grid container with a list of items, filling columns
+    in turn, from a starting location.
     """
 
     def __init__(
@@ -189,12 +205,12 @@ class ByColumn(AutoFit):
             The total number of columns in the grid
         total_rows : int
             The total number of rows in the grid
-        *items_to_move : 
+        *items_to_move :
             One or more global Ids.
          : TYPE
             DESCRIPTION.
 
-       """
+        """
         super().__init__(
             column_index,
             row_index,
@@ -229,10 +245,13 @@ class BulkOperationResult:
     def is_failed(self):
         return not self.is_ok()
 
+    def __str__(self):
+        return f"Succeeded: {self.is_ok()}: Result JSON: {self.data}"
+
 
 class Container:
     """
-     Base class of all Container types
+    Base class of all Container types
     """
 
     @classmethod
@@ -303,7 +322,7 @@ class Container:
 
 class ListContainer(Container):
     """
-     A ListContainer is an ordered container of unlimited capacity
+    A ListContainer is an ordered container of unlimited capacity
     """
 
     def __init__(self, list_container: dict):
@@ -315,14 +334,18 @@ class ListContainer(Container):
 
     def capacity(self) -> int:
         """
-        Unlimited capacity
+        Unlimited capacity, returns sys.maxsize
         """
         return sys.maxsize
 
+    def __repr__(self):
+        return f"{self.__class__.__name__}, id={self.data['globalId']!r},storesContainers={self.accept_containers()},\
+storesSubsamples={self.accept_subsamples}"
+
 
 class Workbench(Container):
     """
-      A specialised Container holding currently active samples and containers.
+    A specialised Container holding currently active samples and containers.
     """
 
     def __init__(self, workbench: dict):
@@ -400,6 +423,10 @@ class GridContainer(Container):
                     rc.append((col, row))
         return rc
 
+    def __repr__(self):
+        return f"{self.__class__.__name__} id={self.data['globalId']!r}, storesContainers={self.accept_containers()},\
+ storesSubsamples={self.accept_subsamples()}, percent_full={self.percent_full():.2f}"
+
 
 class SearchFilter:
     def __init__(
@@ -411,6 +438,20 @@ class SearchFilter:
         if owned_by is not None and len(owned_by) > 0:
             self.data["ownedBy"] = owned_by
 
+    def __str__(self):
+        return f"{str(self.data)}"
+
+    def __repr__(self):
+        return f"{self.__class__.__name__}({self.data['deletedItems']!r}, '{self.data['ownedBy']!r}')"
+
+    def __eq__(self, other):
+        if not isinstance(other, self.__class__):
+            return False
+        return (
+            self.deleted_item_filter == other.deleted_item_filter
+            and self.data["ownedBy"] == other.data["ownedBy"]
+        )
+
 
 class ResultType(Enum):
     SAMPLE = 1
@@ -423,7 +464,11 @@ class Id:
     """
     Supports integer or string representation of a globalId or
     numeric ID or a dict / object representation of an Inventory item.
-    If a dict is passed, it must have 'id' and 'globalId' properties
+    If a dict is passed, it must have 'id' and 'globalId' properties.
+
+    Two Ids are equal in 2 cases:
+        - if their prefix and id are both equal
+        - if only IDs are equal, and neither prefix is defined
     """
 
     Pattern = r"([A-Z]{2})?\d+"
@@ -432,13 +477,13 @@ class Id:
         "IC": "CONTAINER",
         "SS": "SUBSAMPLE",
         "SA": "SAMPLE",
-        "IT": "TEMPLATE",
+        "IT": "SAMPLE_TEMPLATE",
     }
     PREFIX_TO_API = {
         "IC": "containers",
         "SS": "subSamples",
         "SA": "samples",
-        "IT": "templates",
+        "IT": "sampleTemplates",
     }
 
     def __init__(self, value: Union[int, str, dict, Container, Workbench, Sample]):
@@ -476,6 +521,31 @@ class Id:
                 f"Could not interpet {value} as an identifiable Inventory item."
             )
 
+    def __repr__(self):
+        rc = str(self.id)
+        if hasattr(self, "prefix"):
+            rc = "'" + self.prefix + rc + "'"
+        return f"{self.__class__.__name__}({rc})"
+
+    def __str__(self):
+        rc = str(self.id)
+        if hasattr(self, "prefix"):
+            rc = self.prefix + rc
+        return rc
+
+    def __eq__(self, o):
+        if not isinstance(o, self.__class__):
+            return False
+        if self.id != o.id:
+            return False
+        pref_s = hasattr(self, "prefix")
+        pref_o = hasattr(o, "prefix")
+        if (pref_s and pref_o and self.prefix == o.prefix) or (
+            not pref_s and not pref_o
+        ):
+            return True
+        return False
+
     def as_id(self) -> int:
         return self.id
 
@@ -523,6 +593,12 @@ class TemperatureUnit(Enum):
 
 
 class StorageTemperature:
+    """
+    Value object that stores  degrees and units.
+    Two temperatures are considered equal if they differ by less
+     than 1 part in 1e4
+    """
+
     def __init__(
         self, degrees: float, units: TemperatureUnit = TemperatureUnit.CELSIUS
     ):
@@ -532,15 +608,47 @@ class StorageTemperature:
     def _toDict(self) -> dict:
         return {"unitId": self.units.value, "numericValue": self.degrees}
 
+    def __repr__(self):
+        return f"{self.__class__.__name__} ({self.degrees}, {self.units!r})"
+
+    def __str__(self):
+        return f"{self.degrees} {self.units}"
+
+    def __eq__(self, o):
+        if not isinstance(o, self.__class__):
+            return False
+        return (
+            math.isclose(self.degrees, o.degrees, rel_tol=1e-4)
+            and self.units == o.units
+        )
+
 
 class Quantity:
-    def __init__(self, value: float, units: qu.QuantityUnit):
+    """
+    Two quantities are considered equal if they differ by less than 1 part in 1e4
+    """
+
+    def __init__(self, value: float, units: dict):
         self.value = value
         self.units = units
 
     def _toDict(self) -> dict:
         return {"numericValue": self.value, "unitId": self.units["id"]}
 
+    def __repr__(self):
+        return f"{self.__class__.__name__} ({self.value}, '{self.units['label']}')"
+
+    def __str__(self):
+        return f"{self.value} {self.units['label']}"
+
+    def __eq__(self, o):
+        if not isinstance(o, self.__class__):
+            return False
+        return (
+            math.isclose(self.value, o.value, rel_tol=1e-4)
+            and self.units["id"] == o.units["id"]
+        )
+
 
 class ExtraField:
     """
@@ -555,8 +663,18 @@ class ExtraField:
     ):
         self.data = {"name": name, "type": fieldType.value, "content": content}
 
+    def __repr__(self):
+        return f"{self.__class__.__name__} ({self.data['name']!r}, {self.data['type']!r},\
+{self.data['content']!r})"
+
 
 class InventoryClient(ClientBase):
+    """
+    Wrapper around RSpace Inventory API.
+    Enables creation, searching, altering and deleting containers, samples,
+    subsamples and templates.
+    """
+
     API_VERSION = "v1"
 
     def _get_api_url(self):
@@ -572,6 +690,8 @@ class InventoryClient(ClientBase):
         name: str,
         tags: Optional[str] = None,
         description: Optional[str] = None,
+        sample_template_id=None,
+        fields=None,
         extra_fields: Optional[Sequence] = [],
         storage_temperature_min: StorageTemperature = None,
         storage_temperature_max: StorageTemperature = None,
@@ -584,6 +704,9 @@ class InventoryClient(ClientBase):
         Creates a new sample with a mandatory name, optional attributes
         If no template id is specified, the default template will be used,
         whose quantity is measured as a volume.
+
+        Note that including files to attach to Attachment fields is not supported
+        by this method.
         """
         data = self._set_core_properties(name, tags, description, extra_fields)
         if storage_temperature_min is not None:
@@ -596,6 +719,10 @@ class InventoryClient(ClientBase):
             data["newSampleSubSamplesCount"] = subsample_count
         if total_quantity is not None:
             data["quantity"] = total_quantity._toDict()
+        if sample_template_id is not None:
+            data["templateId"] = sample_template_id
+        if fields is not None:
+            data["fields"] = fields
         ## fail early
         if attachments is not None:
             if not isinstance(attachments, list):
@@ -676,9 +803,10 @@ class InventoryClient(ClientBase):
     def _do_simple_list(self, endpoint, pagination, sample_filter):
         if sample_filter is not None:
             pagination.data.update(sample_filter.data)
-        self.serr(f"pg is {pagination.data}")
         return self.retrieve_api_results(
-            f"/{endpoint}", request_type="GET", params=pagination.data,
+            f"/{endpoint}",
+            request_type="GET",
+            params=pagination.data,
         )
 
     def stream_samples(
@@ -715,22 +843,21 @@ class InventoryClient(ClientBase):
             pagination.data.update(sample_filter.data)
         return self._stream("containers", pagination)
 
-    def rename(self, sample_id: Union[str, dict], new_name: str) -> dict:
+    def rename(self, item_id: Union[str, dict], new_name: str) -> dict:
         """
+        Renames an inventory item
         Parameters
         ----------
-            id : Id  of item to rename
+            id : Global Id of item, or a dict representation of the item to rename
             new_name : str The new name.
         Returns
         -------
             dict : The updated item
         """
-        s_id = Id(sample_id)
+        s_id = Id(item_id)
         endpoint = s_id.get_api_endpoint()
         return self.retrieve_api_results(
-            f"/{endpoint}/{s_id.as_id()}",
-            request_type="PUT",
-            params={"name": new_name},
+            f"/{endpoint}/{s_id.as_id()}", request_type="PUT", params={"name": new_name}
         )
 
     def delete_sample(self, sample_id: Union[int, str]):
@@ -767,7 +894,8 @@ class InventoryClient(ClientBase):
         Parameters
         ----------
         - inventory_item : str
-            Global id or dictionary of a sample (SA...), Subsample (SS...) or Container (IC...)
+            Global id or dictionary of a sample (SA...), Subsample (SS...) Container (IC...), or SampleField (SF...)
+            If the item is a SampleField id, then the field must be of type 'Attachment'
         - file : an open file
             An open file stream.
 
@@ -805,12 +933,13 @@ class InventoryClient(ClientBase):
         quantity_per_subsample : float, optional
             The quantity per subsample If not set, the whole subsample will
             be split equally. Use this parameter to set a smaller quantity per subsample.
-   
+
         Returns
         -------
         A list of split subsamples
 
         """
+
         def _do_call(ss_id, params):
             return self.retrieve_api_results(
                 f"/subSamples/{ss_id.as_id()}/actions/split",
@@ -874,7 +1003,7 @@ class InventoryClient(ClientBase):
         Parameters
         ----------
         global_id : str
-            Global id  of template,sample, subsample or container or aict containing global_id
+            Global id  of template,sample, subsample or container or dict containing global_id
         new_name : optional new name of the copy
 
         Returns
@@ -885,7 +1014,8 @@ class InventoryClient(ClientBase):
         id_to_copy = Id(item_to_duplicate)
         endpoint = id_to_copy.get_api_endpoint()
         rc = self.retrieve_api_results(
-            f"/{endpoint}/{id_to_copy.as_id()}/actions/duplicate", request_type="POST",
+            f"/{endpoint}/{id_to_copy.as_id()}/actions/duplicate",
+            request_type="POST",
         )
         if new_name is not None:
             rc = self.rename(rc, new_name)
@@ -943,7 +1073,9 @@ class InventoryClient(ClientBase):
             raise ValueError("Supplied id is not a subsamples")
         data = {"content": note}
         return self.retrieve_api_results(
-            f"/subSamples/{ss_id.as_id()}/notes", request_type="POST", params=data,
+            f"/subSamples/{ss_id.as_id()}/notes",
+            request_type="POST",
+            params=data,
         )
 
     def get_workbenches(self) -> Sequence[dict]:
@@ -958,6 +1090,20 @@ class InventoryClient(ClientBase):
         result = self.retrieve_api_results("/workbenches")
         return [wb for wb in result["containers"]]
 
+    def _configure_parent_container_post(self, location, data):
+        is_wb = False
+        if location == "t":
+            data["removeFromParentContainerRequest"] = True
+        elif location == "w":
+            is_wb = True
+        elif isinstance(location, int) or not is_wb:
+            parent_id = Id(location)
+            if not parent_id.is_container(True):
+                raise ValueError("Id must be that of a container")
+            data["parentContainers"] = [{"id": parent_id.as_id()}]
+        else:
+            raise TypeError("location must be 'w', 't' or a container id or global Id")
+
     def create_list_container(
         self,
         name: str,
@@ -966,12 +1112,14 @@ class InventoryClient(ClientBase):
         extra_fields: Optional[Sequence] = [],
         can_store_containers: bool = True,
         can_store_subsamples: bool = True,
+        location: Union[str, int] = "t",
     ) -> dict:
 
         data = self._set_core_properties(name, tags, description, extra_fields)
         data["cType"] = "LIST"
         data["canStoreContainers"] = can_store_containers
         data["canStoreSubsamples"] = can_store_subsamples
+        self._configure_parent_container_post(location, data)
 
         container = self.retrieve_api_results(
             "/containers", request_type="POST", params=data
@@ -992,6 +1140,7 @@ class InventoryClient(ClientBase):
         extra_fields: Optional[Sequence] = [],
         can_store_containers: bool = True,
         can_store_subsamples: bool = True,
+        location: Union[int, str] = "t",
     ) -> dict:
 
         data = self._set_core_properties(name, tags, description, extra_fields)
@@ -1000,6 +1149,8 @@ class InventoryClient(ClientBase):
         data["canStoreSubsamples"] = can_store_subsamples
         data["gridLayout"] = {"columnsNumber": column_count, "rowsNumber": row_count}
 
+        self._configure_parent_container_post(location, data)
+
         container = self.retrieve_api_results(
             "/containers", request_type="POST", params=data
         )
@@ -1014,7 +1165,9 @@ class InventoryClient(ClientBase):
         )
 
     def add_items_to_list_container(
-        self, target_container_id: Union[str, int], *item_ids: str,
+        self,
+        target_container_id: Union[str, int],
+        *item_ids: str,
     ) -> list:
         """
         Adds 1 or more items to a list container
@@ -1047,7 +1200,9 @@ class InventoryClient(ClientBase):
         for item_id in item_ids:
             id_ob = Id(item_id)
             if not id_ob.is_movable():
-                raise ValueError(f"Item to move '{item_id}' must be a container or subsample")
+                raise ValueError(
+                    f"Item to move '{item_id}' must be a container or subsample"
+                )
             valid_item_ids.append(id_ob)
 
         return self._do_add_to_list_container(valid_item_ids, id_target)
@@ -1251,6 +1406,194 @@ class InventoryClient(ClientBase):
         """
         return self.retrieve_api_results(f"/listOfMaterials/{lom_id}")
 
+    def create_sample_template(self, sample_template_post: dict):
+        """
+        Creates a new SampleTemplate. Use  TemplateBuilder to create the
+        template data structure required as sample_template_post parameter.
+
+        Parameters
+        ----------
+        sample_template_post : A Dict
+            A Dictionary of the SampleTemplate definition to post.
+
+        Returns
+        -------
+        Dict
+            The newly created template.
+
+        """
+        return self.retrieve_api_results(
+            "/sampleTemplates", request_type="POST", params=sample_template_post
+        )
+
+    def get_sample_template_by_id(self, sample_template_id: Union[str, int]) -> dict:
+        """
+        Gets a full sampleTemplate information by id or global id
+        Parameters
+        ----------
+        id : Union[int, str]
+            An integer ID e.g 1234 or a global ID e.g. IT1234
+        Returns
+        -------
+        dict
+            A full description of one sample template
+        """
+        s_id = Id(sample_template_id)
+        return self.retrieve_api_results(f"/sampleTemplates/{s_id.as_id()}")
+
+    def delete_sample_template(self, sample_template_id: Union[int, str]) -> None:
+        """
+        Parameters
+        ----------
+        sample_template_id : Union[int, str]
+            A integer id, or a string id or global ID of the template to delete
+
+        Returns
+        -------
+        None.
+
+        """
+        id_to_delete = Id(sample_template_id)
+        self.doDelete("sampleTemplates", id_to_delete.as_id())
+
+    def set_sample_template_icon(self, sample_template_id: Union[int, str], file):
+        """
+        Parameters
+        ----------
+        sample_template_id : Union[int, str]
+            The ID of the template to add the icon too.
+        file : an open File
+            An icon or image to help identify the template in listings.
+
+        Returns
+        -------
+        The updated SampleTemplate, with an iconId set.
+
+        """
+        st_id = Id(sample_template_id)
+        headers = self._get_headers()
+        response = requests.post(
+            f"{self._get_api_url()}/sampleTemplates/{st_id.as_id()}/icon",
+            files={"file": file},
+            headers=headers,
+        )
+        return self._handle_response(response)
+
+    def get_sample_template_icon(
+        self, sample_template_id: Union[int, str], icon_id: int, outfile
+    ):
+        """
+        Downloads the Sample Template's icon
+
+        Parameters
+        ----------
+        sample_template_id : Union[int, str]
+            The id of the SampleTemplate.
+        icon_id : int
+            A numeric ID of the icon.
+        outfile : string
+            A  path to a writable file to store the downloaded icon.
+
+        Returns
+        -------
+        void, no return value
+
+        """
+        st_id = Id(sample_template_id)
+        url_base = self._get_api_url()
+        return self.download_link_to_file(
+            f"{url_base}/sampleTemplates/{st_id.as_id()}/icon/{icon_id}", outfile
+        )
+
+    def list_sample_templates(
+        self, pagination: Pagination = Pagination(), search_filter: SearchFilter = None
+    ):
+        """
+        Paginated listing of SampleTemplates, optionally filtering by username (owner) or deletion status
+
+        Parameters
+        ----------
+        pagination : Pagination, optional
+            The default is Pagination().
+        search_filter : SearchFilter, optional
+            The default is None.
+
+        Returns
+        -------
+        A standard SearchResult with 'totalHits' attribute and a list of 'templates' with basic information
+        about each template.
+
+        """
+        return self._do_simple_list("sampleTemplates", pagination, search_filter)
+
+    def restore_sample_template(self, sample_template_id: Union[int, str]) -> dict:
+        """
+        Restores a deleted sample template so it will appear in the template listings and
+        be usable to create new samples.
+        If the template is not in a deleted state, this action has no effect.
+        Parameters
+        ----------
+        sample_template_id : Union[int, str]
+            The id of the deleted sample to restore.
+        Returns
+        -------
+        dict
+            The updated template.
+
+        """
+        id_to_restore = Id(sample_template_id)
+        return self.retrieve_api_results(
+            f"/sampleTemplates/{id_to_restore.as_id()}/restore",
+            request_type="PUT",
+        )
+
+    def transfer_sample_template_owner(
+        self, sample_template_id: Union[int, str], new_owner: str
+    ):
+        """
+        Transfers the sample template to the new owner
+        Parameters
+        ----------
+        sample_template_id : Union[int, str]
+            The  id of the sample template to transfer
+        new_owner : str
+            The username of the new owner
+
+        Returns
+        -------
+        dict
+            The updated sample template.
+
+        """
+        st_id = Id(sample_template_id)
+        return self._do_transfer_owner("sampleTemplates", st_id, new_owner)
+
+    def transfer_sample_owner(self, sample_id: Union[int, str], new_owner: str):
+        """
+        Transfers the sample  to the new owner
+        Parameters
+        ----------
+        sample_id : Union[int, str]
+            The  id of the sample  to transfer
+        new_owner : str
+            The username of the new owner
+
+        Returns
+        -------
+        dict
+            The updated sample.
+
+        """
+        sample_id = Id(sample_id)
+        return self._do_transfer_owner("samples", sample_id, new_owner)
+
+    def _do_transfer_owner(self, endpoint, item_id, new_owner):
+        return self.retrieve_api_results(
+            f"/{endpoint}/{item_id.as_id()}/actions/changeOwner",
+            request_type="PUT",
+            params={"owner": {"username": new_owner}},
+        )
+
     def barcode(
         self,
         global_id: Union[str, dict],
@@ -1262,7 +1605,7 @@ class InventoryClient(ClientBase):
         Parameters
         ----------
         global_id : Union[str, dict]
-        
+
         barcode_type:
              The default is Barcode.BARCODE.
 
